# -*- coding: utf-8 -*-
"""
This file shows how we anonymized the data from the post-questionnaire
Since we cannot provide the non-anonymized data, you cannot run this script yourself.

Required files:
    - anonymization_key.xlsx
    - sensitive_post_questionnaire_data_to_remove_human_inv.xlsx
    - Qualtrics_Exports/Human_Inv_Post_Questionnaire_Final_2024_03_25_final.csv
    
Created files:
    - clean_and_anonymize_postquestionnaire_data_output.txt
    - postquestionnaire_anonym.csv

Author: Nele Albers
Date: December 2024
"""

import pandas as pd


if __name__ == "__main__":
    
    # Name of log file
    outputfile_name = "clean_and_anonymize_postquestionnaire_data_output.txt"


    # Load post-questionnaire data
    df_post = pd.read_csv("Qualtrics_Exports/Human_Inv_Post_Questionnaire_Final_2024_03_25_final.csv")


    # Load anonymization key
    df_key = pd.read_excel("anonymization_key.xlsx")


    # Anonymize data
    df_post = pd.merge(df_post, df_key, on = "PROLIFIC_PID")
    df_post = df_post.drop(columns=["PROLIFIC_PID", "Prolific_ID"])


    # Remove columns that we cannot publish
    df_post = df_post.drop(columns=["StartDate", 
                                    "EndDate", 
                                    "Status",
                                    "Progress", 
                                    "Duration (in seconds)",
                                    "RecordedDate", 
                                    "ResponseId",
                                    "UserLanguage", 
                                    "DistributionChannel",
                                    "Q_RecaptchaScore"])


    # Remove empty submissions
    length_before = len(df_post)
    df_post = df_post.dropna(subset=["Effort_1"])
    df_post = df_post.reset_index(drop=True)
    length_after = len(df_post)
    print("Number of submissions removed for being empty:", 
          length_before - length_after)
    with open(outputfile_name, 'w') as f:
        f.write("Number of submissions removed for being empty: " + str(length_before - length_after) + "\n")


    # One person has two finished submissions. We then keep the first recorded one, so the one with the lower index
    df_post_person2_indices = df_post[df_post["rand_id"] == "P1047"].index.to_list()
    print("Indices of person with two completed submissions:", df_post_person2_indices)
    length_before = len(df_post)
    df_post = df_post.drop(index=max(df_post_person2_indices))
    df_post = df_post.reset_index(drop=True)
    length_after = len(df_post)
    print("Number of submissions removed from person who had two completed submissions:", 
          length_before - length_after)
    with open(outputfile_name, 'a') as f:
        f.write("Indices of person with two completed submissions: " + str(df_post_person2_indices) + "\n")
        f.write("Number of submissions removed from person who had two completed submissions: " + str(length_before - length_after) + "\n")
    
    
    # Other people might have one completed and one incomplete submission
    # In that case, we keep the last recorded submission, because incomplete submissions are only recorded after 
    # a few hours and therefore later than completed submissions
    df_post["duplicated_last"] = df_post.duplicated(["rand_id"], keep='last')
    length_before = len(df_post)
    df_post = df_post[df_post["duplicated_last"]==False]
    df_post = df_post.reset_index(drop=True)
    length_after = len(df_post)
    print("Number of submissions removed for multiple submissions:", length_before - length_after)
    with open(outputfile_name, 'a') as f:
        f.write("Number of submissions removed for multiple submissions: " + str(length_before - length_after) + "\n")


    # Remove data from people who did not finish enoug of the questionnaire to have at least 3 passed AC
    length_before = len(df_post)
    df_post = df_post.dropna(subset=["Attention_Check_2", "Attention_Check_1"], thresh = 1)
    df_post = df_post.reset_index(drop=True)
    length_after = len(df_post)
    print("Number of submissions removed for not taking enough of the questionnaire to have at least 3 passed attention checks:", 
          length_before - length_after)
    with open(outputfile_name, 'a') as f:
        f.write("Number of submissions removed for not taking enough of the questionnaire to have at least 3 passed attention checks: " + str(length_before - length_after) + "\n")


    # Remove data from people failing more than 1 attention check
    length_before = len(df_post)
    df_post["passed_ac1"] = [1 if df_post.iloc[i]["PA_Identity_Attention_Check"] in["1", "2"] else 0 for i in range(len(df_post))]
    df_post["passed_ac2"] = [1 if df_post.iloc[i]["Attention_Check_1"] == "-2" else 0 for i in range(len(df_post))]
    df_post["passed_ac3"] = [1 if df_post.iloc[i]["Attention_Check_2"] == "3" else 0 for i in range(len(df_post))]
    # One person added spaces after "Color", which we also decided to accept
    df_post["passed_ac4"] = [1 if df_post.iloc[i]["AC3"] in ["color", "colour", "Color", "Colour"] or df_post.iloc[i]["rand_id"] == "P305" else 0 for i in range(len(df_post))]
    df_post["summed_ac"] = df_post["passed_ac1"] + df_post["passed_ac2"] + df_post["passed_ac3"] + df_post["passed_ac4"]
    df_post = df_post[df_post["summed_ac"] >= 3]
    df_post = df_post.reset_index(drop=True)
    length_after = len(df_post)
    print("Number of submissions removed for not passing enough AC:", length_before - length_after)
    print("Remaining number of submissions:", length_after)
    with open(outputfile_name, 'a') as f:
        f.write("Number of submissions removed for failing 2 or more attention checks: " + str(length_before - length_after) + "\n")
        f.write("Remaining number of submissions: " + str(length_after))


    # Remove sensitive phrases from activity experience
    # There were no sensitive phrases in the other free-text responses in the post-questionnaire
    df_sens = pd.read_excel("sensitive_post_questionnaire_data_to_remove_human_inv.xlsx")
    sens_list = df_sens["phrases_to_remove"].to_list()
    for i in range(len(df_post)):
        for phrase in sens_list:
            df_post.at[i, "Activity_experience"] = df_post.at[i, "Activity_experience"].replace(phrase, "[...]")


    # Remove auxiliary columnes we no longer need
    df_post = df_post.drop(columns=["duplicated_last", 
                                    "passed_ac1", 
                                    "passed_ac2", 
                                    "passed_ac3",
                                    "passed_ac4", 
                                    "summed_ac",
                                    "VERB_ING"])
    
    
    # Remove columns we do not need and do not want to publish for this paper
    df_post = df_post.drop(columns=["Weekly_Exercise",
                                    "Smoking_Freq",
                                    "Last_smoke",
                                    "Vaping_Freq",
                                    "Last_vape",
                                    "HLA",
                                    "HLB",
                                    "NA",
                                    "NB",
                                    "AAS",
                                    "AU",
                                    "PF",
                                    "AL",
                                    "AS",
                                    "APP",
                                    "UAA",
                                    "AE",
                                    "UE",
                                    "UT",
                                    "UAL",
                                    "AA",
                                    "AC",
                                    "AI",
                                    "AT",
                                    "SP",
                                    "IIS",
                                    "AEI",
                                    "UEP",
                                    "UAI",
                                    "Support_read_why",
                                    "user_satisfaction_1",
                                    "user_satisfaction_2",
                                    "user_satisfaction_3",
                                    "user_satisfaction_4",
                                    "user_satisfaction_5",
                                    "user_satisfaction_6",
                                    "user_satisfaction_7",
                                    "user_satisfaction_8",
                                    "P_with_without_R_1",
                                    "P_with_without_E",
                                    "P_effect_R_1",
                                    "P_effect_E",
                                    "P_accountability_R_1",
                                    "P_engagement_R_1"])


    # Save anonymized data
    df_post.to_csv("postquestionnaire_anonym.csv", index=False)
