# -*- coding: utf-8 -*-
"""
Compute different optimal policies using the return likelihood as reward
function and the features selected for the effort-based reward. The resulting files
are stored in the folder "Intermediate_Results_Return_Effortfeatures."

Required files:
    - Data/data_rl_samples_abstracted[0, 1, 2][3, 2, 2].csv

Author: Nele Albers
Date: June 2024
    
"""
from copy import deepcopy
from numpy import nan  # need this for pd.eval
import numpy as np
import pandas as pd
import pickle

import compute_dynamics_feat_sel as dyn
import optimal_policy_computations as pol_comp


PATH = "Intermediate_Results_Return_Effortfeatures/" # pre-fix for path for storing results

NUM_ACTIONS = 2

FEAT_SEL = [0, 1, 2]  # Indices of selected state features
FEATURES_WITH_LEARNED_TRANSITIONS = [0, 1, 2]
NUM_FEAT = len(FEAT_SEL)  # Number of features
NUM_VALS_PER_FEATURE = [3, 2, 2]

DISCOUNT_FACTOR = 0.85

HUMAN_INV_COSTS = [0, 0.01, 0.02, 0.05, 0.06, 0.09, 0.137, 0.14, 0.18, 0.22]

def solve_and_save(reward_func, transitions,
                   path_to_save):
    """
    Compute optimal policy, V, and q-values and store these.

    Parameters
    ----------
    reward_func : list (num_states x num_actions)
        Reward function
    transitions : list (num_states x num_actions x num_states)
        Transition function
    path_to_save : string
        Path to save the optimal policy, V, and q-values.

    Returns
    -------
    None

    """
    num_states = len(reward_func)
    # Optimal policy based on the entire model
    optimal_policy_full, trans_func, V = pol_comp.get_optimal_policy_full(NUM_ACTIONS,
                                                                          reward_func, 
                                                                          transitions, 
                                                                          DISCOUNT_FACTOR)


    print("Number of states with human support:", sum(optimal_policy_full))

    
    # Compute Q-values based on the optimal policy, using the value function V
    q_vals = np.zeros((num_states, NUM_ACTIONS))
    for s_idx in range(num_states):
        for a in range(NUM_ACTIONS):
            q_vals[s_idx, a] = reward_func[s_idx, a] + DISCOUNT_FACTOR * sum(trans_func[a][s_idx, :] * V)

    with open(PATH + "_qvals_" + path_to_save, 'wb') as f:
        pickle.dump(q_vals, f)


if __name__ == "__main__":

    ### Load transition data
    data = pd.read_csv("Data/data_rl_samples_abstracted" + str(FEAT_SEL) + str(NUM_VALS_PER_FEATURE) + ".csv",
                       converters={'s0': eval, 's1': eval})
    
    data_train = data.copy(deep=True)
    # Mean effort spent on activities
    return_mean = data_train["dropout_response"].mean()


    ### Combinations of values for features
    states = [[i, j, k] for i in range(NUM_VALS_PER_FEATURE[0]) for j in range(NUM_VALS_PER_FEATURE[1]) for k in range(NUM_VALS_PER_FEATURE[2])]
    num_states = len(states)
    
    
    path_to_save =  str(str(FEAT_SEL[0]) + str(FEAT_SEL[1]) + str(FEAT_SEL[2]) + "_" + str(DISCOUNT_FACTOR) + "_" + str(NUM_VALS_PER_FEATURE))

    # When we have less than 10 values for a feature value combination for an action,
    # we balance the reward with the mean.
    reward_func, transitions = dyn.compute_dynamics(data_train[["s0", "s1", "a", "dropout_response"]].values.tolist(),
                                                    return_mean, 
                                                    FEAT_SEL,
                                                    NUM_ACTIONS, 
                                                    NUM_VALS_PER_FEATURE,
                                                    features_with_learned_transitions = FEATURES_WITH_LEARNED_TRANSITIONS,
                                                    comp_rewards = True,
                                                    input_reward_lower_bound = -5,
                                                    input_reward_upper_bound = 5)


    with open(PATH + "_reward_func_" + path_to_save, "wb") as f:
        pickle.dump(reward_func, f)
    with open(PATH + "_trans_func_" + path_to_save, "wb") as f:
        pickle.dump(transitions, f)
        

    for cost in HUMAN_INV_COSTS:

        path_to_save_cost = path_to_save
        if cost != 0:
            path_to_save_cost += "_cost" + str(cost)
        
        # Add cost to reward function
        reward_func_cost = deepcopy(reward_func)
        for state in range(num_states):
            reward_func_cost[state, 1] -= cost
            
        # Compute optimal policy and Q-values
        solve_and_save(reward_func_cost, transitions, path_to_save_cost)
