# -*- coding: utf-8 -*-
"""
Functions for optimal policy computations, including with auxiliary rewards.

Author: Nele Albers
Date: September 2023.
"""

import mdptoolbox
import numpy as np
from scipy import sparse


NUM_ACTIVITIES = 53  # total number of activities (preparatory + persuasive)
NUM_PREPARATORY_ACTIVITIES = 44
NUM_PREP_CLUSTERS = 5  # number of clusters of preparatory activities


def get_optimal_policy_full(num_actions,
                            reward_func, 
                            transitions,
                            discount_factor,
                            verbose = True):
    """
    Computes the full optimal policy.

    Parameters
    ----------
    num_actions: int
        Number of actions
    reward_func : np.array
        Reward function
    transitions : np.array
        Transition function
    discount_factor : float
        Discount factor for optimal policy computation
    verbose : boolean
        Whether to print results from iterations

    Returns
    -------
    list
        Optimal policy
    sparse matrix
        Transition function
    tuple
        V* per state
    """
    NUM_STATES = len(reward_func)

    transition_func = []
    for action in range(num_actions):
        transition_func.append(sparse.lil_matrix((NUM_STATES, NUM_STATES),
                               dtype="float64"))
        
    for s_idx in range(NUM_STATES):
        # Transitions for all actions
        for action in range(num_actions):
            for s2_idx in range(NUM_STATES):
                transition_func[action][s_idx, s2_idx] = transitions[s_idx, action, s2_idx]

    for action in range(num_actions):
        transition_func[action] = transition_func[action].tocsr()

    m = mdptoolbox.mdp.ValueIterationGS(transition_func, reward_func, 
                                        discount = discount_factor, 
                                        epsilon=0.001, max_iter=1000)
    if verbose:
        m.setVerbose()
    m.run()
    
    return m.policy, transition_func, m.V


def compute_reward_func_base_to_aux_statespace(reward_func, states, num_feat, 
                                               states_aux, num_actions,
                                               cost):
    """
    Compute the base reward function for the state space that includes also the
    state features used for the auxiliary reward computation.

    Parameters
    ----------
    reward_func : list (num_base_states x num_actions)
        Base reward function.
    states : list
        Base states
    num_feat : int
        Number of features in base states.
    states_aux : list
        States including the features needed for the auxiliary rewards.
    num_actions : int
        Number of actions
    cost : float (>= 0)
        Cost of giving human support

    Returns
    -------
    reward_func_base_for_aux : np-array (num_states x num_actions)
        Reward function for the larger state space.

    """
    
    num_states_aux = len(states_aux)

    reward_func_base_for_aux = np.zeros((num_states_aux, num_actions))
    for state_1_idx, state_1 in enumerate(states_aux):
        base_state = states.index(state_1[0:num_feat])
        # a = 0
        reward_func_base_for_aux[state_1_idx, 0] = reward_func[base_state, 0]
        # a = 1
        reward_func_base_for_aux[state_1_idx, 1] = reward_func[base_state, 1] - cost
        
    return reward_func_base_for_aux


def compute_reward_func_user_weighted(reward_func, 
                                      states, 
                                      num_feat,
                                      states_aux, 
                                      num_actions, 
                                      feat_index_priority, 
                                      num_vals_feat_priority,
                                      feat_index_request, 
                                      num_vals_feat_request,
                                      feat_index_waitingtime, 
                                      num_vals_feat_waitingtime,
                                      weight_prognosis,
                                      aux_weight_priority, 
                                      aux_weight_request, 
                                      aux_weight_waitingtime, 
                                      aux_weight_worststate,
                                      cost):
    """
    Compute the reward function based on the weights given by users in the
    post-questionnaire of our study.

    Parameters
    ----------
    reward_func : list (num_base_states x num_actions)
        Base reward function.
    states : list
        Base states
    num_feat : int
        Number of features in base states.
    states_aux : list
        States including the features needed for the auxiliary rewards.
    num_actions : int
        Number of actions
    feat_index_priority : int
        State index for the priority
    num_vals_feat_priority : int
        Number of values for the priority feature
    feat_index_request : int
        State index for the requesting human support feature
    num_vals_feat_request : int
        Number of values for the requesting human support state feature
    feat_index_waitingtime : int
        State index for the waiting time feature
    num_vals_feat_waitingtime : int
        Number of values for the waiting time state feature
    weight_prognosis : float
        Weight for the basis reward
    aux_weight_priority : float
        Weight for the priority auxiliary reward
    aux_weight_request : float
        Weight for the autonomy auxiliary reward
    aux_weight_waitingtime : flaot
        Weight for the waiting time auxiliary reward
    aux_weight_worststate : float
        Weight for the worst state auxiliary reward
    cost : float
        Cost of giving human support

    Returns
    -------
    reward_func_with_aux : np-array (num_states x num_actions)
        Reward function that includes the auxiliary reward.

    """
    
    # The weights may not sum to 1, because part of the action choice might also be done randomly
    weight_sum = weight_prognosis + aux_weight_priority + aux_weight_request + aux_weight_waitingtime + aux_weight_worststate
    
    # Highest reward when not giving human support
    max_rew_0 = max(reward_func[:, 0])
    # Lowest reward when not giving human support
    min_rew_0 = min(reward_func[:, 0])

    # Highest reward when giving human support
    max_rew_1 = max(reward_func[:, 1])
    # Lowest reward when giving human support
    min_rew_1 = min(reward_func[:, 1])
    
    num_states_aux = len(states_aux)

    reward_func_with_aux = np.zeros((num_states_aux, num_actions))
    for state_1_idx, state_1 in enumerate(states_aux):
        base_state_idx = states.index(state_1[0:num_feat])
        ### a = 0
        # Also scale this reward the same way as we scale the base reward for a = 1
        reward_func_with_aux[state_1_idx, 0] = (reward_func[base_state_idx, 0] - min_rew_1)/(max_rew_1 - min_rew_1)
        ### a = 1
        # We also scale the prognosis reward so that the lowest value is 0 and the highest 1
        # so that the values are similar to the auxiliary rewards and the weighting is done properly
        reward_prognosis = (reward_func[base_state_idx, 1] - min_rew_1)/(max_rew_1 - min_rew_1)
        # Auxiliary rewards
        aux_reward_priority = state_1[feat_index_priority]/(num_vals_feat_priority - 1)
        aux_reward_worststate = (max_rew_0 - reward_func[base_state_idx, 0])/(max_rew_0 - min_rew_0)
        aux_reward_request = state_1[feat_index_request]/(num_vals_feat_request - 1)
        aux_reward_waitingtime = state_1[feat_index_waitingtime]/(num_vals_feat_waitingtime - 1)
        # Combine rewards
        comb_reward = (weight_prognosis * reward_prognosis + aux_weight_priority * aux_reward_priority \
            + aux_weight_request * aux_reward_request + aux_weight_waitingtime * aux_reward_waitingtime \
                + aux_weight_worststate * aux_reward_worststate) / weight_sum
        reward_func_with_aux[state_1_idx, 1] = comb_reward - cost
        
    return reward_func_with_aux


def compute_reward_func_worst_state(reward_func, states, num_feat, 
                                    states_aux, num_actions, aux_weight,
                                    cost):
    """
    Compute the reward function that includes an auxiliary reward based on the
    worst state, whereby the worst state is the one with the lowest reward when
    no human support is given.

    Parameters
    ----------
    reward_func : list (num_base_states x num_actions)
        Base reward function.
    states : list
        Base states
    num_feat : int
        Number of features in base states.
    states_aux : list
        States including the features needed for the auxiliary rewards.
    num_actions : int
        Number of actions
    aux_weight : float (0 - 1)
        Weight given to the auxiliary reward
    cost : float
        Cost of giving human support

    Returns
    -------
    reward_func_with_aux : np-array (num_states x num_actions)
        Reward function that includes the auxiliary reward.

    """
    
    num_states_aux = len(states_aux)

    # Highest reward when not giving human support
    max_rew_0 = max(reward_func[:, 0])
    # Lowest reward when not giving human support
    min_rew_0 = min(reward_func[:, 0])
    
    # Highest reward when giving human support
    max_rew_1 = max(reward_func[:, 1])
    # Lowest reward when giving human support
    min_rew_1 = min(reward_func[:, 1])
    
    # Reward function
    reward_func_with_aux = np.zeros((num_states_aux, num_actions))
    for state_1_idx, state_1 in enumerate(states_aux):
        base_state_idx = states.index(state_1[0:num_feat])
        # a = 0 (i.e., no human support)
        # Also scale this reward the same way as we scale the base reward for a = 1
        reward_func_with_aux[state_1_idx, 0] = (reward_func[base_state_idx, 0] - min_rew_1)/(max_rew_1 - min_rew_1)
        # a = 1 (i.e., human support)
        # also scale the base reward so that the weights are appropriate
        base_reward = (reward_func[base_state_idx, 1] - min_rew_1)/(max_rew_1 - min_rew_1)
        aux_reward = (max_rew_0 - reward_func[base_state_idx, 0])/(max_rew_0 - min_rew_0)
        comb_reward = aux_weight * aux_reward + (1 - aux_weight) * base_reward
        reward_func_with_aux[state_1_idx, 1] = comb_reward - cost
        
    return reward_func_with_aux


def compute_reward_func_priority(reward_func, states, num_feat, 
                                 states_aux, num_actions, aux_weight,
                                 feat_index_priority, num_vals_feat_priority,
                                 cost):
    """
    Compute the reward function that includes an auxiliary reward based on a
    priority.

    Parameters
    ----------
    reward_func : list (num_base_states x num_actions)
        Base reward function.
    states : list
        Base states
    num_feat : int
        Number of features in base states.
    states_aux : list
        States including the features needed for the auxiliary rewards.
    num_actions : int
        Number of actions
    aux_weight : float (0 - 1)
        Weight given to the auxiliary reward
    feat_index_priority : int
        State index for the priority
    num_vals_feat_priority : int
        Number of values for the priority feature
    cost : float
        Cost of giving human support

    Returns
    -------
    reward_func_with_aux : np-array (num_states x num_actions)
        Reward function that includes the auxiliary reward.

    """
    
    num_states_aux = len(states_aux)
    
    # Highest reward when giving human support
    max_rew_1 = max(reward_func[:, 1])
    # Lowest reward when giving human support
    min_rew_1 = min(reward_func[:, 1])

    reward_func_with_aux = np.zeros((num_states_aux, num_actions))
    for state_1_idx, state_1 in enumerate(states_aux):
        base_state = states.index(state_1[0:num_feat])
        # a = 0
        # Also scale this reward the same way as we scale the base reward for a = 1
        reward_func_with_aux[state_1_idx, 0] = (reward_func[base_state, 0] - min_rew_1)/(max_rew_1 - min_rew_1)
        # a = 1
        base_reward = (reward_func[base_state, 1] - min_rew_1)/(max_rew_1 - min_rew_1)
        aux_reward = state_1[feat_index_priority]/(num_vals_feat_priority - 1)
        comb_reward = aux_weight * aux_reward + (1 - aux_weight) * base_reward
        reward_func_with_aux[state_1_idx, 1] = comb_reward - cost
        
    return reward_func_with_aux


def compute_reward_func_request(reward_func, states, num_feat, 
                                states_aux, num_actions, aux_weight,
                                feat_index_request, num_vals_feat_request,
                                cost):
    """
    Compute the reward function that includes an auxiliary reward based on requesting
    human support.

    Parameters
    ----------
    reward_func : list (num_base_states x num_actions)
        Base reward function.
    states : list
        Base states
    num_feat : int
        Number of features in base states.
    states_aux : list
        States including the features needed for the auxiliary rewards.
    num_actions : int
        Number of actions
    aux_weight : float (0 - 1)
        Weight given to the auxiliary reward
    feat_index_request : int
        State index for the requesting human support feature
    num_vals_feat_request : int
        Number of values for the requesting human support state feature
    cost : float
        Cost of giving human support

    Returns
    -------
    reward_func_with_aux : np-array (num_states x num_actions)
        Reward function that includes the auxiliary reward.

    """
    num_states_aux = len(states_aux)
    
    # Highest reward when giving human support
    max_rew_1 = max(reward_func[:, 1])
    # Lowest reward when giving human support
    min_rew_1 = min(reward_func[:, 1])
    
    reward_func_with_aux = np.zeros((num_states_aux, num_actions))
    for state_1_idx, state_1 in enumerate(states_aux):
        base_state = states.index(state_1[0:num_feat])
        # a = 0
        # Also scale this reward the same way as we scale the base reward for a = 1
        reward_func_with_aux[state_1_idx, 0] = (reward_func[base_state, 0] - min_rew_1)/(max_rew_1 - min_rew_1)
        # a = 1
        base_reward = (reward_func[base_state, 1] - min_rew_1)/(max_rew_1 - min_rew_1)
        aux_reward = state_1[feat_index_request]/(num_vals_feat_request - 1)
        comb_reward = aux_weight * aux_reward + (1 - aux_weight) * base_reward
        reward_func_with_aux[state_1_idx, 1] = comb_reward - cost
    
    return reward_func_with_aux


def compute_reward_func_waitingtime(reward_func, states, num_feat, 
                                    states_aux, num_actions, aux_weight,
                                    feat_index_waitingtime, num_vals_feat_waitingtime,
                                    cost):
    """
    Compute the reward function that includes an auxiliary reward based on a
    priority.

    Parameters
    ----------
    reward_func : list (num_base_states x num_actions)
        Base reward function.
    states : list
        Base states
    num_feat : int
        Number of features in base states.
    states_aux : list
        States including the features needed for the auxiliary rewards.
    num_actions : int
        Number of actions
    aux_weight : float (0 - 1)
        Weight given to the auxiliary reward
    feat_index_waitingtime : int
        State index for the waiting time feature
    num_vals_feat_waitingtime : int
        Number of values for the waiting time state feature
    cost : float
        Cost of giving human support

    Returns
    -------
    reward_func_with_aux : np-array (num_states x num_actions)
        Reward function that includes the auxiliary reward.

    """
    num_states_aux = len(states_aux)
    
    # Highest reward when giving human support
    max_rew_1 = max(reward_func[:, 1])
    # Lowest reward when giving human support
    min_rew_1 = min(reward_func[:, 1])
    
    reward_func_with_aux = np.zeros((num_states_aux, num_actions))
    for state_1_idx, state_1 in enumerate(states_aux):
        base_state = states.index(state_1[0:num_feat])
        # a = 0
        # Also scale this reward the same way as we scale the base reward for a = 1
        reward_func_with_aux[state_1_idx, 0] = (reward_func[base_state, 0] - min_rew_1)/(max_rew_1 - min_rew_1)
        # a = 1
        base_reward = (reward_func[base_state, 1] - min_rew_1)/(max_rew_1 - min_rew_1)
        aux_reward = state_1[feat_index_waitingtime]/(num_vals_feat_waitingtime - 1)
        comb_reward = aux_weight * aux_reward + (1 - aux_weight) * base_reward
        reward_func_with_aux[state_1_idx, 1] = comb_reward - cost
    
    return reward_func_with_aux


def get_transitions_with_aux(transitions, states, states_aux, num_actions, 
                             num_vals_feat_priority, 
                             num_vals_feat_waitingtime,
                             num_sessions_per_waitingtime):
    """
    Compute the transition function with the additional state features that
    determine auxiliary rewards.

    Parameters
    ----------
    transitions : array (num_base_states x num_actions x num_base_states)
        Learned transitions for the base states
    states : list
        Base states
    states_aux : list
        States including the additional features for the auxiliary rewards
    num_actions : int
        Number of actions
    num_vals_feat_priority : int
        Number of values for the priority state feature
    num_vals_feat_waitingtime : int
        Number of values for the waiting time state feature
    num_sessions_per_waitingtime : int
        Number of sessions per value of the waiting time state feature

    Returns
    -------
    transitions_with_aux : np.array (num_states x num_actions x num_states)
        Transition function with the additional state features

    """
    
    num_states_aux = len(states_aux)

    transitions_with_aux = np.zeros((num_states_aux, num_actions, num_states_aux))
    for state_1_idx, state_1 in enumerate(states):
        for state_2_idx, state_2 in enumerate(states):
            # the priority always stays unchanged
            for p1 in range(num_vals_feat_priority):
                for w1 in range(num_vals_feat_waitingtime):
                    state_1_aux = state_1 + [p1, w1]
                    state_1_aux_idx = states_aux.index(state_1_aux)
                    # last possible waiting time value
                    if w1 == num_vals_feat_waitingtime - 1:
                        # will stay in waiting time if a = 0
                        state_2_aux = state_2 + [p1, w1]
                        state_2_aux_idx = states_aux.index(state_2_aux)
                        transitions_with_aux[state_1_aux_idx, 0, state_2_aux_idx] = transitions[state_1_idx, 0, state_2_idx]
                        # will move to 0 waiting time if a = 1
                        state_2_aux = state_2 + [p1, 0]
                        state_2_aux_idx = states_aux.index(state_2_aux)
                        transitions_with_aux[state_1_aux_idx, 1, state_2_aux_idx] = transitions[state_1_idx, 1, state_2_idx]
                    else:
                        # will move to next waiting time with certain probability if a = 0
                        prob_stay = (num_sessions_per_waitingtime - 1)/num_sessions_per_waitingtime
                        state_2_aux = state_2 + [p1, w1]
                        state_2_aux_idx = states_aux.index(state_2_aux)
                        transitions_with_aux[state_1_aux_idx, 0, state_2_aux_idx] = prob_stay * transitions[state_1_idx, 0, state_2_idx]
                        state_2_aux = state_2 + [p1, w1 + 1]
                        state_2_aux_idx = states_aux.index(state_2_aux)
                        transitions_with_aux[state_1_aux_idx, 0, state_2_aux_idx] = (1 - prob_stay) * transitions[state_1_idx, 0, state_2_idx]
                        # will move to 0 waiting time if a = 1
                        state_2_aux = state_2 + [p1, 0]
                        state_2_aux_idx = states_aux.index(state_2_aux)
                        transitions_with_aux[state_1_aux_idx, 1, state_2_aux_idx] = transitions[state_1_idx, 1, state_2_idx]
    
    return transitions_with_aux
