# -*- coding: utf-8 -*-
"""
Abstract the state features after the feature selection and add data from
people's Prolific profiles and prescreening questionnaires.

Required files:
    - Data/Activity_Difficulty_Data/Merged_Data_Difficulty_Ratings_Activities.csv
    - Data/prescreening_questionnaire_anonym.csv
    - Data/prolific_profile_anonym.csv
    - RQ1/Data/data_rl_samples.csv
    - all_states
    - all_states_sessions
    
Created files:
    - DataOverview_RQ2_RQ3/Data/data_rl_samples_abstracted[0, 1, 2][3, 2, 2].csv
    - DataOverview_RQ2_RQ3/Data/all_abstract_states_with_session.csv
    - create_abstracted_rl_samples_output.txt
    

Author: Nele Albers
Date: June 2024
"""

from copy import deepcopy
from numpy import nan  # need this for pd.eval
import numpy as np
import pandas as pd
import pickle


# State features users were asked about in each session
STATE_FEATURE_NAMES_ASKED = ["state_importance", 
                             "state_selfefficacy",
                             "state_humansupport",
                             "state_energy"]
NUM_STATE_FEATURES_ASKED = len(STATE_FEATURE_NAMES_ASKED)
STATE_INDEX_DIFFICULTY_RATING = NUM_STATE_FEATURES_ASKED


NUM_ACTIVITIES = 37  # Number of preparatory activities


# Number of values we want to have per state feature in the end
# order: importance, self-efficacy, appreciation, energy, difficulty, phase
# (Values for not-selected features do not really matter)
NUM_ABS_VALS_PER_FEATURE = [3, 2, 2, 3, 2, 2]


FEAT_SEL = [0, 1, 2]  # Indices of selected features
STATES = [[i, j, k] for i in range(NUM_ABS_VALS_PER_FEATURE[FEAT_SEL[0]]) for j in range(NUM_ABS_VALS_PER_FEATURE[FEAT_SEL[1]]) for k in range(NUM_ABS_VALS_PER_FEATURE[FEAT_SEL[2]])]


def eval_with_nan(x):
    lst = pd.eval(x)
    return lst


def get_abstract_states(df, all_states,
                        num_abs_vals_per_feature, basis_state_feature_names,
                        state_index_difficulty_rating, difficulty_cutoffs,
                        state_name, outputfile_name):
    """
    Compute the abstract states.

    Parameters
    ----------
    df : dataframe
        Dataframe with non-abstracted states
    all_states : list
        All states (i.e., also ones that are not part of RL-samples), used
        to compute abstract states
    num_abs_vals_per_feature : list
        Number of values we want to have per state feature in the end.
    basis_state_feature_names : list
        Number of original state features
    state_index_difficulty_rating : int
        Position of activity difficulty rating in state.
    difficulty_cutoffs : list
        Cutoffs computed for the difficulty rating.
    state_name: string
        Column name of the state to be abstracted.
    outputfile_name: string
        Name of the output file for logging. Leave empty to not log.
    

    Returns
    -------
    df_abs : dataframe
       Same as input dataframe but with abstracted states.

    """

    df_abs = df.copy(deep=True)
    num_rows = len(df_abs)

    # Number of basis state features
    num_basis_state_features = len(basis_state_feature_names)

    # We also need a deep copy of the state lists
    for row in range(len(df_abs)):
        df_abs.at[row, state_name] = deepcopy(df_abs.at[row, state_name])


    # We compute an abstract version of the state features based on the cutoffs
    # If we have partial data on a state for a session, we set the remaining missing
    # values equal to what the median value would be mapped to
    for feat in range(num_basis_state_features):
        feat_cutoffs = np.nanpercentile([all_states[row][feat] for row in range(len(all_states))], [j/num_abs_vals_per_feature[feat] * 100 for j in range(1, num_abs_vals_per_feature[feat] + 1)])
        if len(outputfile_name) > 0:
            print("Cutoffs for feature " + basis_state_feature_names[feat] + ":", np.round(feat_cutoffs, 2))
            with open(outputfile_name, 'a') as f:
                f.write("Cutoffs for feature " + basis_state_feature_names[feat] + ": " + str(np.round(feat_cutoffs, 2)) + "\n")
        for row in range(num_rows):
            df_abs = get_abstract_val(df_abs, feat_cutoffs,
                                      state_name, row, feat)


    # And we do the same thing for the activity difficulty rating
    for row in range(num_rows):
        df_abs = get_abstract_val(df_abs, difficulty_cutoffs,
                                  state_name, row, state_index_difficulty_rating)


    return df_abs


def get_abstract_val(df_abs, cutoffs,
                     state_name, row, 
                     state_feat_index):
    """
    Get abstract value for one state for one feature.

    Parameters
    ----------
    df_abs : dataframe
        Dataframe we are using to compute abstracted state features
    cutoffs : list
        Value cutoffs to use for the abstraction for this state feature.
    state_name : string
        Column name of state to be abstracted.
    row : int
        Row index for dataframe indicating the row we are working on.
    state_feat_index : int
        State feature index in the state.

    Returns
    -------
    df_abs : dataframe
        Same as input dataframe but with one feature of one state abstracted.

    """
    # Number of possible values for abstract feature
    num_abs_vals_feat_curr = len(cutoffs)

    for cutoff_idx, cutoff in enumerate(cutoffs):
        if df_abs.at[row, state_name][state_feat_index] <= cutoff:
            df_abs.at[row, state_name][state_feat_index] = cutoff_idx
            break
    # If we have missing data, we set the missing value equal to the abstract value
    # the median is mapped to
    if np.isnan(df_abs.loc[row, state_name][state_feat_index]):
        df_abs.loc[row, state_name][state_feat_index] = int(np.floor((num_abs_vals_feat_curr - 1)/2))

    return df_abs


if __name__ == "__main__":

    # Log file
    outputfile_name = "create_abstracted_rl_samples_output.txt"

    ### Load data
    # Load RL-samples with non-abstracted state features
    df_rl_samples = pd.read_csv("RQ1/Data/data_rl_samples.csv",
                                converters={'s0': eval_with_nan, 's1': eval_with_nan})

    # All states (i.e., also those that are not part of transition samples from session 1)
    with open("all_states", 'rb') as f:
        all_states = pickle.load(f)
    # And corresponding sessions
    with open("all_states_sessions", 'rb') as f:
        all_states_sessions = pickle.load(f)
    
    # Data from Prolific profiles and prescreening questionnarie
    df_prolific = pd.read_csv("Data/prolific_profile_anonym.csv")[["Age", 
                                                                   "Gender", 
                                                                   "Highest education level completed",
                                                                   "rand_id"]]
    df_prescreening = pd.read_csv("Data/prescreening_questionnaire_anonym.csv")[["PIV_Smoking_Freq",
                                                                                 "Vaping_Freq",
                                                                                 "ttm_smoking_vaping",
                                                                                 "ttm_pa",
                                                                                 "rand_id",
                                                                                 "Weekly_Exercise"]]
    
    
    ### Compute difficulty rating cutoffs
    df_difficulty_ratings = pd.read_csv("Data/Activity_Difficulty_Data/Merged_Data_Difficulty_Ratings_Activities.csv")
    difficulty_ratings_means = [df_difficulty_ratings[str(i) + "_Ease"].mean() for i in range(NUM_ACTIVITIES)]
    difficulty_cutoffs = np.nanpercentile(difficulty_ratings_means, [j/NUM_ABS_VALS_PER_FEATURE[STATE_INDEX_DIFFICULTY_RATING] * 100 for j in range(1, NUM_ABS_VALS_PER_FEATURE[STATE_INDEX_DIFFICULTY_RATING] + 1)])
    print("Activity difficulty cutoffs:", np.round(difficulty_cutoffs, 2))
    with open(outputfile_name, 'w') as f:
        f.write("Activity difficulty cutoffs: " + str(np.round(difficulty_cutoffs, 2)) + "\n")


    ### Compute abstracted state feature values
    with open(outputfile_name, 'a') as f:
        f.write("***Abstracting RL samples***\n")
    df_rl_samples_abs = df_rl_samples
    for state_name in ["s0", "s1"]:
        with open(outputfile_name, 'a') as f:
            f.write(state_name + ":\n")
        df_rl_samples_abs = get_abstract_states(df_rl_samples_abs, all_states,
                                                NUM_ABS_VALS_PER_FEATURE,
                                                STATE_FEATURE_NAMES_ASKED,
                                                STATE_INDEX_DIFFICULTY_RATING, 
                                                difficulty_cutoffs,
                                                state_name = state_name,
                                                outputfile_name = outputfile_name)
    
    
        # Also create separate columns for abstract state features
        df_rl_samples_abs[[state_name + '_abs_imp', state_name + "_abs_se", 
                           state_name + "_abs_hs", state_name + "_abs_energy", 
                           state_name + "_abs_diff", state_name + "_abs_session"]] = pd.DataFrame(df_rl_samples_abs[state_name].tolist(), 
                                                                                                  index= df_rl_samples_abs.index)


    # And we merge the Prolific profile data and prescreening questionnaire data
    df_rl_samples_abs = pd.merge(df_rl_samples_abs, df_prolific, on = "rand_id")
    df_rl_samples_abs = pd.merge(df_rl_samples_abs, df_prescreening, on = "rand_id")


    # Save data
    df_rl_samples_abs.to_csv("DataOverview_RQ2_RQ3/Data/data_rl_samples_abstracted" + str(FEAT_SEL) + str([NUM_ABS_VALS_PER_FEATURE[f] for f in FEAT_SEL]) + ".csv", 
                             index=False)


    ### Also compute abstracted version of all states
    # Create dataframe
    df_all_states = pd.DataFrame()
    df_all_states["state"] = all_states
    df_all_states["session"] = all_states_sessions
    # Separate columns for all state features
    df_all_states[['state_imp', "state_se", "state_hs", 
                   "state_energy", "state_diff", "state_session"]] = pd.DataFrame(df_all_states.state.tolist(), index= df_all_states.index)
    # Compute abstract states
    df_all_states = get_abstract_states(df_all_states, all_states,
                                        NUM_ABS_VALS_PER_FEATURE,
                                        STATE_FEATURE_NAMES_ASKED,
                                        STATE_INDEX_DIFFICULTY_RATING, 
                                        difficulty_cutoffs,
                                        state_name = "state",
                                        outputfile_name = "")
    df_all_states.to_csv("DataOverview_RQ2_RQ3/Data/all_abstract_states_with_session.csv", 
                         index=False)
