# -*- coding: utf-8 -*-
"""
Create the non-abstracted RL-samples.

Required files:
    - Data/Activity_Difficulty_Data/Merged_Data_Difficulty_Ratings_Activities.csv
    - Data/sessionsdata_anonym.csv
    
Output files:
    - all_states
    - all_states_sessions
    - DataOverview_RQ2_RQ3/Data/all_states
    - DataOverview_RQ2_RQ3/Data/data_rl_samples.csv
    - RQ1/Data/data_rl_samples.csv
    - merge_data_output.txt
    

Author: Nele Albers
Date: June 2024
"""

from copy import deepcopy
import numpy as np
import pandas as pd
import pickle


NUM_SESSIONS = 5
# State features users were asked about in each session
STATE_FEATURE_NAMES_ASKED = ["state_importance", 
                             "state_selfefficacy",
                             "state_humansupport",
                             "state_energy"]
NUM_STATE_FEATURES_ASKED = len(STATE_FEATURE_NAMES_ASKED)
NUM_ACTIVITIES = 37  # Number of preparatory activities


if __name__ == "__main__":


    # Name of log file
    outputfile_name = "merge_data_output.txt"


    with open(outputfile_name, 'w') as f:
        f.write("***Merging data***\n")


    ### Load data
    df_session = pd.read_csv("Data/sessionsdata_anonym.csv")
    df_difficulty_ratings = pd.read_csv("Data/Activity_Difficulty_Data/Merged_Data_Difficulty_Ratings_Activities.csv")


    ### Compute mean difficulty rating values for the activities
    difficulty_ratings_means = [df_difficulty_ratings[str(i) + "_Ease"].mean() for i in range(NUM_ACTIVITIES)]


    ### Collect transition samples
    all_first_states  = []
    all_next_states = []
    session_nums = []
    all_efforts = []
    all_actions = []
    all_ids = []
    all_activities = []
    all_dropouts = []

    all_states = []  # all states to compute means
    all_states_sessions = []  # session numbers corresponding to all_states


    for session in range(1, NUM_SESSIONS):


        # All IDs of people with at least an answer to state_importance for the next session
        # (which is the first state feature)
        all_session_ids = list(set(df_session[(df_session["session_num"] == session + 1) & (df_session["response_type"] == "state_importance")]["rand_id"].to_list()))


        for i in all_session_ids:

            # Get data from this person
            df_id = df_session[df_session["rand_id"] == i]


            # Get the activity assigned to the person
            activity_curr = int(df_id[(df_id["session_num"] == session) & (df_id["response_type"] == "activity_new_index")]["response_value"])


            # Get the first state
            state = []
            for state_feature_name in STATE_FEATURE_NAMES_ASKED:
                state.append(int(df_id[(df_id["session_num"] == session) & (df_id["response_type"] == state_feature_name)]["response_value"]))
            # Append activity difficulty rating
            state.append(difficulty_ratings_means[activity_curr])
            # Append the session (starting at index 0)
            state.append(session - 1)


            # Get the next state
            next_state = []
            df_id_session_states = df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"].isin(STATE_FEATURE_NAMES_ASKED))]
            for state_feature_name in STATE_FEATURE_NAMES_ASKED:
                if state_feature_name in df_id_session_states["response_type"].to_list():
                    next_state.append(float(df_id_session_states[df_id_session_states["response_type"] == state_feature_name]["response_value"]))
            # Fill up with nan if we do not have data on all state features
            num_features_obtained = len(next_state)
            for f in range(len(next_state), NUM_STATE_FEATURES_ASKED):
                if f == num_features_obtained:
                    print("Next state with missing feature values. Only have data on " + str(len(next_state)) + " feature(s).")
                    with open(outputfile_name, 'a') as f:
                        f.write("Next state with missing feature values. Only have data on " + str(len(next_state)) + " feature(s).\n")
                next_state.append(np.nan)
            # Append next activity difficulty rating if it exists
            df_id_session_activity = df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"] == "activity_new_index")]
            if len(df_id_session_activity) > 0:
                next_state.append(difficulty_ratings_means[int(df_id_session_activity["response_value"].to_list()[0])])
            else:
                print("Next state with missing activity difficulty.")
                with open(outputfile_name, 'a') as f:
                    f.write("Next state with missing activity difficulty.\n")
                next_state.append(np.nan)
            # Append the session (starting at index 0)
            next_state.append(session)


            all_first_states.append(state)
            all_next_states.append(next_state)
            session_nums.append(session)
            all_efforts.append(int(df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"] == "effort")]["response_value"]))
            all_dropouts.append(int(df_id[(df_id["session_num"] == session + 1) & (df_id["response_type"] == "dropout_response")]["response_value"]))
            all_actions.append(int(df_id[(df_id["session_num"] == session) & (df_id["response_type"] == "human_support_after_session")]["response_value"]))
            all_activities.append(activity_curr)
            all_ids.append(i)


            # Add the state to the list of all states for computing feature means
            all_states.append(deepcopy(state))
            all_states_sessions.append(session)  # session with indices 1-5


            # To get all states, we also need to add the next state if that next state is the last state of the person
            # By counting the number of times the person filled in information for the state_importance state feature,
            # we know how many sessions that person completed (almost) fully (some state features might be missing
            # from the last state)
            df_id_state_importance = df_id[df_id["response_type"] == "state_importance"]
            if session == len(df_id_state_importance) - 1:
                all_states.append(deepcopy(next_state))
                all_states_sessions.append(session + 1)


    ### To get all states, we also need to add the first states of people who do not have data for at least the first state feature in session 2
    # Get all IDs of people with data for at least the first state feature in session 2
    all_ids_arrived_in_session2 = list(set(df_session[(df_session["session_num"] == 2) & (df_session["response_type"] == "state_importance")]["rand_id"].to_list()))
    # IDs of people with data for at least the first state feature in session 1 who do not have data for at least the first state feature in session 2
    all_ids_arrived_in_session1 = list(set(df_session[(df_session["session_num"] == 1) & (df_session["response_type"] == "state_importance")]["rand_id"].to_list()))
    all_ids_arrived_in_session1_but_not_session2 = [i for i in all_ids_arrived_in_session1 if not i in all_ids_arrived_in_session2]
    print("Number of people with data on first state only:", len(all_ids_arrived_in_session1_but_not_session2))
    with open(outputfile_name, 'a') as f:
        f.write("Number of people with data on first state only: " + str(len(all_ids_arrived_in_session1_but_not_session2)) + "\n")


    # Get the first states of those people with no first state feature data in session 2
    for i in all_ids_arrived_in_session1_but_not_session2:

        # Get data from this person
        df_id = df_session[df_session["rand_id"] == i]


        # Get the first state
        first_state = []
        df_id_session_states = df_id[(df_id["session_num"] == 1) & (df_id["response_type"].isin(STATE_FEATURE_NAMES_ASKED))]
        for state_feature_name in STATE_FEATURE_NAMES_ASKED:
            if state_feature_name in df_id_session_states["response_type"].to_list():
                first_state.append(float(df_id_session_states[df_id_session_states["response_type"] == state_feature_name]["response_value"]))
        # Fill up with nan if we do not have data on all state features
        num_features_obtained = len(first_state)
        for f in range(len(first_state), NUM_STATE_FEATURES_ASKED):
            if f == num_features_obtained:
                print("First state with missing feature values. Only have data on " + str(len(first_state)) + " feature(s).")
                with open(outputfile_name, 'a') as f:
                    f.write("First state with missing feature values. Only have data on " + str(len(first_state)) + " feature(s).\n")
            first_state.append(np.nan)
        # Add activity difficulty rating to state
        activity_first = int(df_id[(df_id["session_num"] == 1) & (df_id["response_type"] == "activity_new_index")]["response_value"])
        first_state.append(difficulty_ratings_means[activity_first])
        # Add the session
        first_state.append(0)


        all_states.append(first_state)
        all_states_sessions.append(1)


    # Save all_states and corresponding sessions
    with open("all_states", "wb") as f:
        pickle.dump(all_states, f)
    with open("DataOverview_RQ2_RQ3/Data/all_states", "wb") as f:
        pickle.dump(all_states, f)
    with open("all_states_sessions", "wb") as f:
        pickle.dump(all_states_sessions, f)


    ### Create dataframe with RL samples
    df_rl_samples = pd.DataFrame()
    df_rl_samples["rand_id"] = all_ids
    df_rl_samples["session"] = session_nums
    df_rl_samples["s0"] = all_first_states
    df_rl_samples["a"] = all_actions
    df_rl_samples["effort"] = all_efforts
    df_rl_samples["s1"] = all_next_states 
    df_rl_samples["activity"] = all_activities
    df_rl_samples["dropout_response"] = all_dropouts


    # And we also want to create separate columns with the state features
    df_rl_samples[['s0_imp', "s0_se", "s0_hs", "s0_energy", "s0_diff", "s0_session"]] = pd.DataFrame(df_rl_samples.s0.tolist(), index= df_rl_samples.index)
    df_rl_samples[['s1_imp', "s1_se", "s1_hs", "s1_energy", "s1_diff", "s1_session"]] = pd.DataFrame(df_rl_samples.s1.tolist(), index= df_rl_samples.index)


    # And we also need a consecutive ID for participants for the models in R
    all_ids = list(set(df_rl_samples["rand_id"].to_list()))
    df_rl_samples["cons_id"] = [all_ids.index(df_rl_samples.at[i, "rand_id"]) for i in range(len(df_rl_samples))]
    
    
    # We also need to create a column that captures how many times a person has 
    # received human feedback by that time
    feedback_counts = []
    for i in range(len(df_rl_samples)):
        df_i = df_rl_samples[df_rl_samples["rand_id"] == df_rl_samples.at[i, "rand_id"]]
        df_i_prev = df_i[df_i["session"] <= df_rl_samples.at[i, "session"]]
        df_i_prev_with_support = df_i_prev[df_i_prev["a"] == 1]
        feedback_counts.append(len(df_i_prev_with_support))
    df_rl_samples["Prev_Feedback_Count"] = feedback_counts


    # Save this dataframe
    df_rl_samples.to_csv("RQ1/Data/data_rl_samples.csv", 
                         index=False)
    df_rl_samples.to_csv("DataOverview_RQ2_RQ3/Data/data_rl_samples.csv", 
                         index=False)
